#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#define MAGIC ((char *) 0x0EFFACEE)
#define shmemDEFMODE 0664

typedef struct {
	char   *magic;
	int     id;
	int     refcount;
	int 	lock;
	key_t   key;
	int     mode;
	size_t  size;
	size_t  real;
} shmem_t, *pshmem_t;

pshmem_t 
_shmem_ptr_(char *p) {
	return (pshmem_t) (p - sizeof(shmem_t));
}

int 
is_pshmem(char *p) {
	return p && (p != (char *) -1) && ((_shmem_ptr_(p))->magic == MAGIC);
}

#define _shmem_c_
#include <shmem.h>
#undef _shmem_c_

char *
shmem_attach(int key) {
	int id;
	pshmem_t s;

	if ((id = shmget((key_t) key, 0, 0)) < 0) return NULL;
	if ((s = (pshmem_t) shmat(id, NULL, 0)) == (pshmem_t) -1) 		  return NULL;
	s->refcount++;
	return ((char *) s + sizeof(shmem_t));
}

int
shmem_detach(char *p) {
	pshmem_t s;
	if (is_pshmem(p)) {
		s = _shmem_ptr_(p);
		s->refcount--; 
		return shmdt((void *) s);
	}
	return 1;
}

char *
shmem_destroy(char *p) {
	pshmem_t s;
	if (is_pshmem(p)) {
		int id;	
		s = _shmem_ptr_(p);
		s->refcount--; 
		id = shmem_id(p);
		shmdt((void *) s);
		shmctl(id, IPC_RMID, NULL);
	}
	return NULL;
}

size_t shmem_sizeof() { return sizeof(shmem_t); }

char *
shmem_new(int key, size_t size, int mode) {
	pshmem_t s;
	int id;
	int rc = 1;
	
	if (!mode) mode = shmemDEFMODE;

	if ((id = shmget((key_t) key,  size + sizeof(shmem_t), mode)) < 0) {
		if ((id = shmget((key_t) key, size + sizeof(shmem_t), IPC_CREAT | mode)) < 0) return NULL;
		rc = 0;
	}
	
	if ((s = (pshmem_t) shmat(id, NULL, 0)) == (pshmem_t) -1) return  NULL;
		
	if (rc) s->refcount++;
	else {
		s->magic	= MAGIC;
		s->id       = id;
		s->key      = (key_t) key;
		s->refcount = 1;
		s->lock     = 0;
		s->mode     = mode;
		s->size     = size;	
		s->real     = size + sizeof(shmem_t);
	} 
	return ((char *) s + sizeof(shmem_t));
}

int
shmem_stat(key_t id, struct shmid_ds *stat) {
	if (!stat) return 1;
	if (shmctl(id, IPC_STAT, stat)) return 2;
	return 0;
}

int
shmem_mode_get(char *p) {
	pshmem_t s;
	struct shmid_ds perm;

	if (!is_pshmem(p)) return 1; 
	s = _shmem_ptr_(p);

	if (shmem_stat(s->id, &perm)) return 2;
	return perm.shm_perm.mode;
}

int
shmem_owner_get(char *p) {
	pshmem_t s;
	struct shmid_ds perm;

	if (!is_pshmem(p)) return 1; 
	s = _shmem_ptr_(p);

	if (shmem_stat(s->id, &perm)) return 2;

	return perm.shm_perm.uid;
}

int 
shmem_group_get(char *p) {
	pshmem_t s;
	struct shmid_ds perm;

	if (!is_pshmem(p)) return 1; 
	s = _shmem_ptr_(p);

	if (shmem_stat(s->id, &perm)) return 2;

	return perm.shm_perm.gid;
}

int
shmem_owner_set(char *p, uid_t uid) {
	pshmem_t s;
	struct shmid_ds perm;

	if (!is_pshmem(p)) return 1; 
	s = _shmem_ptr_(p);

	if (shmem_stat(s->id, &perm)) return 2;

	perm.shm_perm.uid = uid;

	if (shmctl(s->id, IPC_SET,  &perm)) return 3;
	
	return 0;
}
	
int
shmem_mode_set(char *p, int mode) {
	pshmem_t s;
	struct shmid_ds perm;

	if (!is_pshmem(p)) return 1; 
	s = _shmem_ptr_(p);

	if (shmem_stat(s->id, &perm)) return 2;

	perm.shm_perm.mode = mode;

	if (shmctl(s->id, IPC_SET,  &perm)) return 3;
	
	return 0;
}

int
shmem_group_set(char *p, gid_t gid) {
	pshmem_t s;
	struct shmid_ds perm;

	if (!is_pshmem(p)) return 1; 
	s = _shmem_ptr_(p);

	if (shmem_stat(s->id, &perm)) return 2;

	perm.shm_perm.gid = gid;

	if (shmctl(s->id, IPC_SET,  &perm)) return 3;
	
	return 0;
}

int
shmem_id(char *p) { 
	pshmem_t s;
	if (!is_pshmem(p)) return 0; 
	return _shmem_ptr_(p)->id;
}

int
shmem_refcount(char *p) {
	pshmem_t s;
	if (!is_pshmem(p)) return 0; 
	return _shmem_ptr_(p)->refcount;
}

int
shmem_key(char *p) {
	pshmem_t s;
	if (!is_pshmem(p)) return 0; 
	return (int) _shmem_ptr_(p)->key;
}

size_t
shmem_size(char *p) {
	pshmem_t s;
	if (!is_pshmem(p)) return 0; 
	return _shmem_ptr_(p)->size;
}

size_t
shmem_real(char *p) {
	pshmem_t s;
	if (!is_pshmem(p)) return 0; 
	return _shmem_ptr_(p)->real;
}

